@@ -47,16 +47,19 @@ namespace firebuild {
 libconfig::Config * cfg = nullptr;
 
 cstring_view_array ignore_locations {nullptr, 0, 0};
-cstring_view_array system_locations {nullptr, 0, 0};
+cstring_view_array read_only_locations {nullptr, 0, 0};
 
 ExeMatcher* shortcut_allow_list_matcher = nullptr;
 ExeMatcher* dont_shortcut_matcher = nullptr;
 ExeMatcher* dont_intercept_matcher = nullptr;
 ExeMatcher* skip_cache_matcher = nullptr;
+tsl::hopscotch_set<std::string>* shells = nullptr;
+bool ccache_disabled = false;
 /** Store results of processes consuming more CPU time (system + user) in microseconds than this. */
 int64_t min_cpu_time_u = 0;
 int shortcut_tries = 0;
 int64_t max_cache_size = 0;
+uint64_t max_entry_size = 0;
 int quirks = 0;
 
 /**
@@ -157,18 +160,23 @@ static void modify_config(libconfig::Config *cfg, const std::string& str) {
 
   if (append) {
     /* Append scalar value to an existing array. */
-    libconfig::Setting& array = cfg->lookup(name);
-    libconfig::Setting& adding = array.add(type);
-    /* Unfortunately there's no operator= to assign from another Setting. */
-    switch (type) {
-      case libconfig::Setting::TypeString: {
-        std::string x_str(x.c_str());
-        adding = x_str.c_str();
-        break;
+    try {
+      libconfig::Setting& array = cfg->lookup(name);
+      libconfig::Setting& adding = array.add(type);
+      /* Unfortunately there's no operator= to assign from another Setting. */
+      switch (type) {
+        case libconfig::Setting::TypeString: {
+          std::string x_str(x.c_str());
+          adding = x_str.c_str();
+          break;
+        }
+        default:
+          std::cerr << "This type is not supported" << std::endl;
+          exit(EXIT_FAILURE);
       }
-      default:
-        std::cerr << "This type is not supported" << std::endl;
-        exit(EXIT_FAILURE);
+    } catch(libconfig::SettingNotFoundException&) {
+      std::cerr << "Setting not found: " << name << std::endl;
+      exit(EXIT_FAILURE);
     }
   } else if (remove) {
     /* Remove all occurrences of a scalar value from an existing array. */
@@ -241,19 +249,31 @@ static void modify_config(libconfig::Config *cfg, const std::string& str) {
   delete mini_config;
 }
 
-static void init_locations(cstring_view_array* locations, const libconfig::Setting& items) {
+static void init_locations(cstring_view_array* locations, const libconfig::Config *cfg,
+                           const char* locations_setting) {
   cstring_view_array_init(locations);
-  for (int i = 0; i < items.getLength(); i++) {
-    cstring_view_array_append(locations, strdup(items[i].c_str()));
+  try {
+    const libconfig::Setting& items = cfg->getRoot()[locations_setting];
+    for (int i = 0; i < items.getLength(); i++) {
+      cstring_view_array_append(locations, strdup(items[i].c_str()));
+    }
+  } catch(libconfig::SettingNotFoundException&) {
+    /* Configuration setting may be missing. This is OK. */
   }
   cstring_view_array_sort(locations);
 }
 
-static void init_matcher(ExeMatcher **matcher, const libconfig::Setting& items) {
+static void init_matcher(ExeMatcher **matcher, const libconfig::Config *cfg,
+                         const char* matcher_setting) {
   assert(!*matcher);
   *matcher = new ExeMatcher();
-  for (int i = 0; i < items.getLength(); i++) {
-    (*matcher)->add(items[i].c_str());
+  try {
+    const libconfig::Setting& items = cfg->getRoot()["processes"][matcher_setting];
+    for (int i = 0; i < items.getLength(); i++) {
+      (*matcher)->add(items[i].c_str());
+    }
+  } catch(libconfig::SettingNotFoundException&) {
+    /* Configuration setting may be missing. This is OK. */
   }
 }
 
@@ -299,18 +319,50 @@ void read_config(libconfig::Config *cfg, const char *custom_cfg_file,
     }
   }
 
-  assert(FileName::isDbEmpty());
-  init_locations(&ignore_locations, cfg->getRoot()["ignore_locations"]);
-  init_locations(&system_locations, cfg->getRoot()["system_locations"]);
+  if (cfg->exists("max_entry_size")) {
+    libconfig::Setting& max_entry_size_cfg = cfg->getRoot()["max_entry_size"];
+    if (max_entry_size_cfg.isNumber()) {
+      double max_entry_size_mb = max_entry_size_cfg;
+      if (max_entry_size_mb < 0) {
+        /* Fix up negative numbers. */
+        max_entry_size_mb = 0;
+      }
+      max_entry_size = max_entry_size_mb * 1000000;
+    }
+  }
 
-  init_matcher(&shortcut_allow_list_matcher, cfg->getRoot()["processes"]["shortcut_allow_list"]);
+  assert(FileName::isDbEmpty());
+  init_locations(&ignore_locations, cfg, "ignore_locations");
+  init_locations(&read_only_locations, cfg, "read_only_locations");
+  /* The read_only_locations setting used to be called system_locations. */
+  try {
+    const libconfig::Setting& items = cfg->getRoot()["system_locations."];
+    for (int i = 0; i < items.getLength(); i++) {
+      cstring_view_array_append(&read_only_locations, strdup(items[i].c_str()));
+    }
+    cstring_view_array_sort(&read_only_locations);
+  } catch(libconfig::SettingNotFoundException&) {
+    /* Configuration setting may be missing. This is OK. */
+  }
+
+  init_matcher(&shortcut_allow_list_matcher, cfg, "shortcut_allow_list");
   if (shortcut_allow_list_matcher->empty()) {
     delete(shortcut_allow_list_matcher);
     shortcut_allow_list_matcher = nullptr;
   }
-  init_matcher(&dont_shortcut_matcher, cfg->getRoot()["processes"]["dont_shortcut"]);
-  init_matcher(&dont_intercept_matcher, cfg->getRoot()["processes"]["dont_intercept"]);
-  init_matcher(&skip_cache_matcher, cfg->getRoot()["processes"]["skip_cache"]);
+  init_matcher(&dont_shortcut_matcher, cfg, "dont_shortcut");
+  init_matcher(&dont_intercept_matcher, cfg, "dont_intercept");
+  init_matcher(&skip_cache_matcher, cfg, "skip_cache");
+
+  shells = new tsl::hopscotch_set<std::string>();
+  try {
+    libconfig::Setting& shells_cfg = cfg->getRoot()["processes"]["shells"];
+    for (int i = 0; i < shells_cfg.getLength(); i++) {
+      shells->emplace(shells_cfg[i]);
+    }
+  } catch(libconfig::SettingNotFoundException&) {
+    /* Configuration setting may be missing. This is OK. */
+  }
 
   if (cfg->exists("quirks")) {
     const libconfig::Setting& items = cfg->getRoot()["quirks"];
@@ -322,6 +374,8 @@ void read_config(libconfig::Config *cfg, const char *custom_cfg_file,
         quirks |= FB_QUIRK_LTO_WRAPPER;
       } else if (quirk == "ignore-time-queries") {
         quirks |= FB_QUIRK_IGNORE_TIME_QUERIES;
+      } else if (quirk == "ignore-statfs") {
+        quirks |= FB_QUIRK_IGNORE_STATFS;
       } else if (quirk == "guess-file-params") {
         quirks |= FB_QUIRK_GUESS_FILE_PARAMS;
       } else {
@@ -333,27 +387,38 @@ void read_config(libconfig::Config *cfg, const char *custom_cfg_file,
   }
 }
 
+static void export_sorted(const libconfig::Setting& setting,
+                          const std::string env_var_name,
+                          std::map<std::string, std::string>* env) {
+  std::vector<std::string> entries;
+  for (int i = 0; i < setting.getLength(); i++) {
+    entries.emplace_back(setting[i].c_str());
+  }
+
+  if (entries.size() > 0) {
+    std::sort(entries.begin(), entries.end());
+    std::string entries_appended;
+    for (auto entry : entries) {
+      if (entries_appended.length() == 0) {
+        entries_appended.append(entry);
+      } else {
+        entries_appended.append(":" + entry);
+      }
+    }
+    (*env)[env_var_name] = std::string(entries_appended);
+    FB_DEBUG(FB_DEBUG_PROC, " " + env_var_name + "=" + (*env)[env_var_name]);
+  }
+}
+
 static void export_sorted_locations(libconfig::Config *cfg, const char* configuration_name,
                                     const std::string env_var_name,
                                     std::map<std::string, std::string>* env) {
   const libconfig::Setting& root = cfg->getRoot();
-  const libconfig::Setting& locations_setting = root[configuration_name];
-  std::vector<std::string> locations;
-  for (int i = 0; i < locations_setting.getLength(); i++) {
-    locations.emplace_back(locations_setting[i].c_str());
-  }
-  if (locations.size() > 0) {
-    std::sort(locations.begin(), locations.end());
-    std::string locations_appended;
-    for (auto loc : locations) {
-      if (locations_appended.length() == 0) {
-        locations_appended.append(loc);
-      } else {
-        locations_appended.append(":" + loc);
-      }
-    }
-    (*env)[env_var_name] = std::string(locations_appended);
-    FB_DEBUG(FB_DEBUG_PROC, " " + env_var_name + "=" + (*env)[env_var_name]);
+  try {
+    const libconfig::Setting& locations_setting = root[configuration_name];
+    export_sorted(locations_setting, env_var_name, env);
+  } catch(libconfig::SettingNotFoundException&) {
+    /* Configuration setting may be missing. This is OK. */
   }
 }
 
@@ -361,45 +426,70 @@ char** get_sanitized_env(libconfig::Config *cfg, const char *fb_conn_string,
                          bool insert_trace_markers) {
   const libconfig::Setting& root = cfg->getRoot();
 
-  FB_DEBUG(FB_DEBUG_PROC, "Passing through environment variables:");
-
-  const libconfig::Setting& pass_through = root["env_vars"]["pass_through"];
   std::map<std::string, std::string> env;
-  for (int i = 0; i < pass_through.getLength(); i++) {
-    std::string pass_through_env(pass_through[i].c_str());
-    char * got_env = getenv(pass_through_env.c_str());
-    if (got_env != NULL) {
-      env[pass_through_env] = std::string(got_env);
-      FB_DEBUG(FB_DEBUG_PROC, " " + std::string(pass_through_env) + "="
-               + env[pass_through_env]);
+  FB_DEBUG(FB_DEBUG_PROC, "Passing through environment variables:");
+  try {
+    const libconfig::Setting& pass_through = root["env_vars"]["pass_through"];
+    for (int i = 0; i < pass_through.getLength(); i++) {
+      std::string pass_through_env(pass_through[i].c_str());
+      char * got_env = getenv(pass_through_env.c_str());
+      if (got_env != NULL) {
+        env[pass_through_env] = std::string(got_env);
+        FB_DEBUG(FB_DEBUG_PROC, " " + std::string(pass_through_env) + "="
+                 + env[pass_through_env]);
+      }
     }
+    FB_DEBUG(FB_DEBUG_PROC, "");
+  } catch(libconfig::SettingNotFoundException&) {
+    /* Configuration setting may be missing. This is OK. */
   }
-  FB_DEBUG(FB_DEBUG_PROC, "");
 
   FB_DEBUG(FB_DEBUG_PROC, "Setting preset environment variables:");
-  const libconfig::Setting& preset = root["env_vars"]["preset"];
-  for (int i = 0; i < preset.getLength(); i++) {
-    std::string str(preset[i].c_str());
-    size_t eq_pos = str.find('=');
-    if (eq_pos == std::string::npos) {
-      fb_error("Invalid present environment variable: " + str);
-      abort();
-    } else {
-      const std::string var_name = str.substr(0, eq_pos);
-      env[var_name] = str.substr(eq_pos + 1);
-      FB_DEBUG(FB_DEBUG_PROC, " " + var_name + "=" + env[var_name]);
+  try {
+    const libconfig::Setting& preset = root["env_vars"]["preset"];
+    for (int i = 0; i < preset.getLength(); i++) {
+      std::string str(preset[i].c_str());
+      size_t eq_pos = str.find('=');
+      if (eq_pos == std::string::npos) {
+        fb_error("Invalid present environment variable: " + str);
+        abort();
+      } else {
+        const std::string var_name = str.substr(0, eq_pos);
+        env[var_name] = str.substr(eq_pos + 1);
+        if (str == "CCACHE_DISABLE=1") {
+          ccache_disabled = true;
+        }
+        FB_DEBUG(FB_DEBUG_PROC, " " + var_name + "=" + env[var_name]);
+      }
     }
+  } catch(libconfig::SettingNotFoundException&) {
+    /* Configuration setting may be missing. This is OK. */
   }
 
-  export_sorted_locations(cfg, "system_locations", "FB_SYSTEM_LOCATIONS", &env);
+
+  export_sorted_locations(cfg, "read_only_locations", "FB_READ_ONLY_LOCATIONS", &env);
   export_sorted_locations(cfg, "ignore_locations", "FB_IGNORE_LOCATIONS", &env);
 
-  const char *ld_preload_value = getenv("LD_PRELOAD");
-  if (ld_preload_value) {
-    env["LD_PRELOAD"] = LIBFIREBUILD_SO ":" + std::string(ld_preload_value);
-  } else {
-    env["LD_PRELOAD"] = LIBFIREBUILD_SO;
+  try {
+    const libconfig::Setting& locations_setting = root["processes"]["jobserver_users"];
+    export_sorted(locations_setting, "FB_JOBSERVER_USERS", &env);
+  } catch(libconfig::SettingNotFoundException&) {
+    /* Configuration setting may be missing. This is OK. */
   }
+
+  const char *ld_preload_value = getenv(LD_PRELOAD);
+  if (ld_preload_value) {
+    env[LD_PRELOAD] = LIBFIREBUILD_SO ":" + std::string(ld_preload_value);
+  } else {
+    env[LD_PRELOAD] = LIBFIREBUILD_SO;
+  }
+  FB_DEBUG(firebuild::FB_DEBUG_PROC, " " LD_PRELOAD "=" + env[LD_PRELOAD]);
+
+#ifdef __APPLE__
+  env["DYLD_FORCE_FLAT_NAMESPACE"] = "0";
+  FB_DEBUG(firebuild::FB_DEBUG_PROC, " DYLD_FORCE_FLAT_NAMESPACE=" +
+           env["DYLD_FORCE_FLAT_NAMESPACE"]);
+#endif
   env["FB_SOCKET"] = fb_conn_string;
   FB_DEBUG(FB_DEBUG_PROC, " FB_SOCKET=" + env["FB_SOCKET"]);
 
