@@ -18,9 +18,14 @@
 
 #include "firebuild/firebuild.h"
 
+#ifdef __APPLE__
+#include <crt_externs.h>
+#endif
 #include <signal.h>
 #include <getopt.h>
+#ifdef __linux__
 #include <sys/prctl.h>
+#endif
 #include <sys/resource.h>
 #include <sys/socket.h>
 #include <sys/time.h>
@@ -190,7 +195,7 @@ static void accept_ic_conn(const struct epoll_event* event, void *arg) {
 }
 
 static bool running_under_valgrind() {
-  const char *v = getenv("LD_PRELOAD");
+  const char *v = getenv(LD_PRELOAD);
   if (v == NULL) {
     return false;
   } else {
@@ -371,7 +376,7 @@ int main(const int argc, char *argv[]) {
                                                insert_trace_markers);
 
   /* Set up sigchild handler */
-  if (pipe2(sigchild_selfpipe, O_CLOEXEC | O_NONBLOCK) != 0) {
+  if (fb_pipe2(sigchild_selfpipe, O_CLOEXEC | O_NONBLOCK) != 0) {
     firebuild::fb_perror("pipe");
     exit(EXIT_FAILURE);
   }
@@ -382,14 +387,16 @@ int main(const int argc, char *argv[]) {
   sigaction(SIGCHLD, &sa, NULL);
 
   /* Configure epoll */
-  firebuild::epoll = new firebuild::Epoll(EPOLL_CLOEXEC);
+  firebuild::epoll = new firebuild::Epoll();
 
   /* Open listener socket before forking child to always let the child connect */
   listener = create_listener();
   firebuild::epoll->add_fd(listener, EPOLLIN, accept_ic_conn, NULL);
 
+#ifdef __linux__
   /* Collect orphan children */
   prctl(PR_SET_CHILD_SUBREAPER, 1);
+#endif
 
   /* run command and handle interceptor messages */
   if ((child_pid = fork()) == 0) {
@@ -414,7 +421,12 @@ int main(const int argc, char *argv[]) {
       exit(EXIT_FAILURE);
     }
 
+#ifdef __APPLE__
+    *_NSGetEnviron() = env_exec;
+    execvp(argv[optind], argv_exec);
+#else
     execvpe(argv[optind], argv_exec, env_exec);
+#endif
     firebuild::fb_perror("Executing build command failed");
     exit(EXIT_FAILURE);
   } else {
@@ -433,7 +445,8 @@ int main(const int argc, char *argv[]) {
     firebuild::epoll->add_fd(sigchild_selfpipe[0], EPOLLIN, firebuild::sigchild_cb, NULL);
 
     /* Main loop for processing interceptor messages */
-    while (listener >= 0) {
+    /* Runs until the only remaining epoll-monitored fd is the sigchild_selfpipe fd. */
+    while (firebuild::epoll->fds() > 1) {
       /* This is where the process spends its idle time: waiting for an event over a fd, or a
        * sigchild.
        *
@@ -444,6 +457,8 @@ int main(const int argc, char *argv[]) {
 
       /* Process the reported events, if any. */
       firebuild::epoll->process_all_events();
+
+      firebuild::proc_tree->GcProcesses();
     }
 
     /* Finish all top pipes */
@@ -482,20 +497,33 @@ int main(const int argc, char *argv[]) {
       timeradd(&ru_myslf.ru_stime, &ru_chldr.ru_stime, &ru_total.ru_stime);
 
       fprintf(stderr, "\nResource usages, in seconds:\n"
-                      "real           %5ld.%03ld\n"
-                      "user firebuild %5ld.%03ld\n"
-                      "user children  %5ld.%03ld\n"
-                      "user total     %5ld.%03ld\n"
-                      "sys  firebuild %5ld.%03ld\n"
-                      "sys  children  %5ld.%03ld\n"
-                      "sys  total     %5ld.%03ld\n",
-                      diff_time.tv_sec, diff_time.tv_nsec / (1000 * 1000),
-                      ru_myslf.ru_utime.tv_sec, ru_myslf.ru_utime.tv_usec / 1000,
-                      ru_chldr.ru_utime.tv_sec, ru_chldr.ru_utime.tv_usec / 1000,
-                      ru_total.ru_utime.tv_sec, ru_total.ru_utime.tv_usec / 1000,
-                      ru_myslf.ru_stime.tv_sec, ru_myslf.ru_stime.tv_usec / 1000,
-                      ru_chldr.ru_stime.tv_sec, ru_chldr.ru_stime.tv_usec / 1000,
-                      ru_total.ru_stime.tv_sec, ru_total.ru_stime.tv_usec / 1000);
+              "real           %5ld.%03ld\n"
+#ifdef __APPLE__
+              "user firebuild %5ld.%03d\n"
+              "user children  %5ld.%03d\n"
+              "user total     %5ld.%03d\n"
+              "sys  firebuild %5ld.%03d\n"
+              "sys  children  %5ld.%03d\n"
+              "sys  total     %5ld.%03d\n"
+#else
+              "user firebuild %5ld.%03ld\n"
+              "user children  %5ld.%03ld\n"
+              "user total     %5ld.%03ld\n"
+              "sys  firebuild %5ld.%03ld\n"
+              "sys  children  %5ld.%03ld\n"
+              "sys  total     %5ld.%03ld\n"
+#endif
+              "\n"
+              "firebuild's memory usage in MiB:\n"
+              "max. res. set  %9.03f\n",
+              diff_time.tv_sec, diff_time.tv_nsec / (1000 * 1000),
+              ru_myslf.ru_utime.tv_sec, ru_myslf.ru_utime.tv_usec / 1000,
+              ru_chldr.ru_utime.tv_sec, ru_chldr.ru_utime.tv_usec / 1000,
+              ru_total.ru_utime.tv_sec, ru_total.ru_utime.tv_usec / 1000,
+              ru_myslf.ru_stime.tv_sec, ru_myslf.ru_stime.tv_usec / 1000,
+              ru_chldr.ru_stime.tv_sec, ru_chldr.ru_stime.tv_usec / 1000,
+              ru_total.ru_stime.tv_sec, ru_total.ru_stime.tv_usec / 1000,
+              static_cast<double>(ru_myslf.ru_maxrss) / 1024);
     }
 
     if (firebuild::execed_process_cacher->is_gc_needed()) {
@@ -521,7 +549,12 @@ int main(const int argc, char *argv[]) {
   unlink(fb_conn_string);
   rmdir(fb_tmp_dir);
 
+#ifdef FB_EXTRA_DEBUG
+  (void)running_under_valgrind;
+  {
+#else
   if (running_under_valgrind()) {
+#endif
     /* keep Valgrind happy */
     {
       char* env_str;
