@@ -156,8 +156,14 @@ void MessageProcessor::accept_exec_child(ExecedProcess* proc, int fd_conn,
 
       // TODO(rbalint) skip reopening fd if parent's other forked processes closed the fd
       // without writing to it
+      const int jobserver_fd_w = proc->jobserver_fd_w();
       for (inherited_file_t& inherited_file : inherited_files) {
         if (inherited_file.type == FD_PIPE_OUT) {
+          if (inherited_file.fds[0] == jobserver_fd_w
+              && inherited_file.fds.size() == 1) {
+            /* Skip reopening the jobserver pipe. */
+            continue;
+          }
           auto file_fd_old = proc->get_shared_fd(inherited_file.fds[0]);
           auto pipe = file_fd_old->pipe();
           assert(pipe);
@@ -179,7 +185,7 @@ void MessageProcessor::accept_exec_child(ExecedProcess* proc, int fd_conn,
 
           /* Create a new unnamed pipe. */
           int fifo_fd[2];
-          int ret = pipe2(fifo_fd, file_fd->flags() & ~O_ACCMODE);
+          int ret = fb_pipe2(fifo_fd, file_fd->flags() & ~O_ACCMODE);
           (void)ret;
           assert(ret == 0);
           if (epoll->is_added_fd(fifo_fd[0])) {
@@ -271,7 +277,7 @@ static void accept_popen_child(Process* unix_parent, const pending_popen_t *pend
      * accept_exec_child() when reopening the inherited outgoing pipes. */
     FB_DEBUG(FB_DEBUG_PROC, "This is a popen(..., \"r...\") child");
 
-    if (pipe2(down, flags & ~O_ACCMODE) < 0) {
+    if (fb_pipe2(down, flags & ~O_ACCMODE) < 0) {
       assert(0 && "pipe2() failed");
     }
     bump_fd_age(down[0]);
@@ -315,7 +321,7 @@ static void accept_popen_child(Process* unix_parent, const pending_popen_t *pend
      * pipes, as well as the Pipe object handling them. */
     FB_DEBUG(FB_DEBUG_PROC, "This is a popen(..., \"w...\") child");
 
-    if (pipe2(up, flags & ~O_ACCMODE) < 0 || pipe2(down, flags & ~O_ACCMODE) < 0) {
+    if (fb_pipe2(up, flags & ~O_ACCMODE) < 0 || fb_pipe2(down, flags & ~O_ACCMODE) < 0) {
       assert(0 && "pipe2() failed");
     }
     if (epoll->is_added_fd(up[0])) {
@@ -611,7 +617,7 @@ static void proc_ic_msg(const FBBCOMM_Serialized *fbbcomm_buf, uint16_t ack_num,
     case FBBCOMM_TAG_system: {
       auto ic_msg = reinterpret_cast<const FBBCOMM_Serialized_system *>(fbbcomm_buf);
       assert_null(proc->system_child());
-      /* system(cmd) launches a child of argv = ["sh", "-c", cmd] */
+      /* system(cmd) launches a child of argv = ["sh", "-c", "--", cmd], with "--" being optional */
       auto expected_child = new ExecedProcessEnv(proc->pass_on_fds(false), LAUNCH_TYPE_SYSTEM);
       if (ic_msg->has_cmd()) {
         expected_child->set_sh_c_command(ic_msg->get_cmd());
@@ -652,7 +658,7 @@ static void proc_ic_msg(const FBBCOMM_Serialized *fbbcomm_buf, uint16_t ack_num,
 
       int type_flags = ic_msg->get_type_flags();
       auto fds = proc->pass_on_fds(false);
-      /* popen(cmd) launches a child of argv = ["sh", "-c", cmd] */
+      /* popen(cmd) launches a child of argv = ["sh", "-c", "--", cmd], with "--" being optional */
       auto expected_child = new ExecedProcessEnv(fds, LAUNCH_TYPE_POPEN);
       // FIXME what if !has_cmd() ?
       expected_child->set_sh_c_command(ic_msg->get_cmd());
@@ -687,7 +693,7 @@ static void proc_ic_msg(const FBBCOMM_Serialized *fbbcomm_buf, uint16_t ack_num,
       auto ic_msg = reinterpret_cast<const FBBCOMM_Serialized_popen_failed *>(fbbcomm_buf);
       // FIXME what if !has_cmd() ?
       delete(proc->pop_expected_child_fds(
-          std::vector<std::string>({"sh", "-c", ic_msg->get_cmd()}),
+          std::vector<std::string>({"sh", "-c", "--", ic_msg->get_cmd()}),
           nullptr, nullptr, true));
       break;
     }
@@ -742,7 +748,7 @@ static void proc_ic_msg(const FBBCOMM_Serialized *fbbcomm_buf, uint16_t ack_num,
                 /* Pretend that the parent opened the file for writing and not the fork child.
                  * This is not accurate, but the fork child does not exist yet. A parallel
                  * process opening the file for writing would disable shortcutting the same way. */
-                file_name->open_for_writing(proc);
+                file_name->open_for_writing(proc->exec_point());
               }
             }
             break;
@@ -1086,6 +1092,7 @@ static void proc_ic_msg(const FBBCOMM_Serialized *fbbcomm_buf, uint16_t ack_num,
       PFBBA_HANDLE(proc, fchmodat, fbbcomm_buf);
       break;
     }
+#ifdef __linux__
     case FBBCOMM_TAG_memfd_create: {
       PFBBA_HANDLE(proc, memfd_create, fbbcomm_buf);
       break;
@@ -1106,10 +1113,16 @@ static void proc_ic_msg(const FBBCOMM_Serialized *fbbcomm_buf, uint16_t ack_num,
       PFBBA_HANDLE(proc, signalfd, fbbcomm_buf);
       break;
     }
+#endif
     case FBBCOMM_TAG_getrandom: {
       auto ic_msg = reinterpret_cast<const FBBCOMM_Serialized_getrandom *>(fbbcomm_buf);
       const unsigned int flags = ic_msg->get_flags_with_fallback(0);
+#ifdef GRND_RANDOM
       const std::string pathname(flags & GRND_RANDOM ? "/dev/random" : "/dev/urandom");
+#else
+      const std::string pathname("/dev/urandom");
+      (void)flags;
+#endif
       if (!FileName::Get(pathname)->is_in_ignore_location()) {
         proc->exec_point()->disable_shortcutting_bubble_up(
             deduplicated_string("Using " + pathname + " is not allowed").c_str());
@@ -1176,6 +1189,10 @@ static void proc_ic_msg(const FBBCOMM_Serialized *fbbcomm_buf, uint16_t ack_num,
       /* Ignore gethostname. With a local cache it should not make a difference, while
        * in a shared cache the intention is to use cached results from other machines. */
       break;
+    case FBBCOMM_TAG_seccomp:
+      /* Ignore seccomp(). The interposer always returns EINVAL error to keep interception
+       * working. This breaks sandboxing, but builds can run arbitrary commands anyway. */
+      break;
     case FBBCOMM_TAG_fb_debug:
     case FBBCOMM_TAG_fb_error:
     case FBBCOMM_TAG_fchownat:
@@ -1191,9 +1208,12 @@ static void proc_ic_msg(const FBBCOMM_Serialized *fbbcomm_buf, uint16_t ack_num,
       break;
     }
     case FBBCOMM_TAG_gen_call: {
+      auto msg = reinterpret_cast<const FBBCOMM_Serialized_gen_call *>(fbbcomm_buf);
+      const int error = msg->get_error_no_with_fallback(0);
       proc->exec_point()->disable_shortcutting_bubble_up(
-          deduplicated_string(std::string(reinterpret_cast<const FBBCOMM_Serialized_gen_call *>(
-              fbbcomm_buf)->get_call()) + " is not supported").c_str());
+          deduplicated_string(
+              std::string(error == 0 ? "" : "failed") + msg->get_call() + " is not supported"
+              + (error == 0 ? "" : " (error: " + d(error) + ")")).c_str());
       break;
     }
     default: {
@@ -1215,7 +1235,7 @@ void MessageProcessor::ic_conn_readcb(const struct epoll_event* event, void *ctx
   const msg_header * header;
   ProcessDebugSuppressor debug_suppressor(proc);
 
-  if (!(event->events & EPOLLIN)) {
+  if (!Epoll::ready_for_read(event)) {
     FB_DEBUG(FB_DEBUG_COMM, "socket " +
              d_fd(Epoll::event_fd(event)) +
              " hung up (" + d(proc) + ")");
