@@ -111,8 +111,9 @@ static void construct_cached_file_name(const std::string &base,
 
 bool ObjCache::store(const Hash &key,
                      const FBBSTORE_Builder * const entry,
+                     off_t stored_blob_bytes,
                      const FBBFP_Serialized * const debug_key) {
-  TRACK(FB_DEBUG_CACHING, "key=%s", D(key));
+  TRACK(FB_DEBUG_CACHING, "key=%s, stored_blob_bytes=%" PRIoff, D(key), stored_blob_bytes);
 
   if (FB_DEBUGGING(FB_DEBUG_CACHING)) {
     FB_DEBUG(FB_DEBUG_CACHING, "ObjCache: storing entry, key " + d(key));
@@ -155,6 +156,14 @@ bool ObjCache::store(const Hash &key,
   // FIXME Is it faster to ftruncate() the file to the desired size, then mmap,
   // then serialize to the mapped memory, then ftruncate() again to the actual size?
   size_t len = entry->measure();
+  if (stored_blob_bytes + len > max_entry_size) {
+    FB_DEBUG(FB_DEBUG_CACHING,
+             "Could not store entry in cache because it would exceed max_entry_size");
+    free(tmpfile);
+    close(fd_dst);
+    return false;
+  }
+
   char *entry_serial = reinterpret_cast<char *>(malloc(len));
   entry->serialize(entry_serial);
   fb_write(fd_dst, entry_serial, len);
@@ -411,6 +420,10 @@ ObjCache::gc_collect_sorted_obj_timestamp_sizes() {
   return obj_timestamp_sizes;
 }
 
+off_t ObjCache::gc_collect_total_objects_size() {
+  return recursive_total_file_size(base_dir_);
+}
+
 void ObjCache::gc_obj_cache_dir(const std::string& path,
                                 tsl::hopscotch_set<AsciiHash>* referenced_blobs,
                                 off_t* cache_bytes, off_t* debug_bytes,
@@ -454,9 +467,11 @@ void ObjCache::gc_obj_cache_dir(const std::string& path,
           } else if ((debug_postfix = strstr(name, kDebugPostfix))) {
             /* Files for debugging cache entries.*/
             if (FB_DEBUGGING(FB_DEBUG_CACHE)) {
-              char* related_name = reinterpret_cast<char*>(alloca(debug_postfix - name + 1));
-              memcpy(related_name, name, debug_postfix - name);
-              related_name[debug_postfix - name] = '\0';
+              const size_t name_len = debug_postfix - name;
+              assert_cmp(name_len, <, FB_PATH_BUFSIZE);
+              char related_name[FB_PATH_BUFSIZE];
+              memcpy(related_name, name, name_len);
+              related_name[name_len] = '\0';
               struct stat st;
               if (fstatat(dirfd(dir), related_name, &st, 0) == 0) {
                 /* Keeping debugging file that has related object. If the object gets removed
