@@ -21,21 +21,28 @@
 
 #include <fcntl.h>
 #include <signal.h>
+
+#ifdef __linux__
 #include <sys/epoll.h>
 #include <sys/eventfd.h>
+#endif
 #include <sys/ioctl.h>
 #include <sys/mman.h>
+#ifdef __linux__
 #include <sys/signalfd.h>
 #include <sys/socket.h>
 #include <sys/timerfd.h>
+#endif
 #include <sys/types.h>
 #include <unistd.h>
 
+#include <cinttypes>
 #include <utility>
 
 #include "common/firebuild_common.h"
 #include "common/platform.h"
 #include "firebuild/pipe_recorder.h"
+#include "firebuild/config.h"
 #include "firebuild/execed_process.h"
 #include "firebuild/forked_process.h"
 #include "firebuild/execed_process_env.h"
@@ -185,7 +192,7 @@ int Process::handle_pre_open(const int dirfd, const char * const ar_name, const
         "Could not find file name to mark as opened for writing");
     return -1;
   } else {
-    name->open_for_writing(this);
+    name->open_for_writing(this->exec_point());
     return 0;
   }
 }
@@ -200,7 +207,11 @@ int Process::handle_open(const int dirfd, const char * const ar_name, const size
          dirfd, D(ar_name), flags, mode, fd, pre_open_sent, error, D_FD(fd_conn), ack_num);
 
   /* O_TMPFILE is actually multiple bits, 0x410000 */
+#ifdef O_TMPFILE
   const bool o_tmpfile_set = (flags & O_TMPFILE) == O_TMPFILE;
+#else
+  const bool o_tmpfile_set = false;
+#endif
   const FileName* name = get_absolute(dirfd, ar_name, ar_len);
   if (!name) {
     // FIXME don't disable shortcutting if openat() failed due to the invalid dirfd
@@ -386,10 +397,40 @@ int Process::handle_close_range(const unsigned int first, const unsigned int las
   return 0;
 }
 
+
+int Process::handle_dlopen(const char * const absolute_filename,
+                           const size_t absolute_filename_len,
+                           const char * const looked_up_filename,
+                           const size_t looked_up_filename_len,
+                           const bool error, int fd_conn, int ack_num) {
+  if (absolute_filename) {
+    /* When failing to dlopen() a file assume it is not present.
+     * This is a safe assumption for shortcutting purposes, since the cache entry
+     * will require the the file to be missing to shortcut the process and if the file
+     * is missing dlopen() would have failed for sure.
+     */
+    return handle_open(AT_FDCWD, absolute_filename, absolute_filename_len,
+                             O_RDONLY, 0, -1, error ? ENOENT : 0, fd_conn, ack_num, false, false);
+  } else if (looked_up_filename) {
+    /* Failed dlopen() could not find the file on the search path.*/
+    /* TODO(rbalint) allow shortcutting the process and mark the file as missing on all the
+     * search path entries as described in dlopen(3). */
+    (void)looked_up_filename_len;
+    exec_point()->disable_shortcutting_bubble_up("Process failed to dlopen() ", looked_up_filename);
+    return -1;
+  } else {
+    if (error) {
+      exec_point()->disable_shortcutting_bubble_up("Process failed to dlopen() the main program");
+      return -1;
+    }
+  }
+  return 0;
+}
+
 int Process::handle_truncate(const char * const ar_name, const size_t ar_len,
                              const off_t length, const int error) {
   TRACKX(FB_DEBUG_PROC, 1, 1, Process, this,
-         "ar_name=%s, length=%ld, error=%d", D(ar_name), length, error);
+         "ar_name=%s, length=%" PRIoff ", error=%d", D(ar_name), length, error);
 
   const FileName* name = get_absolute(AT_FDCWD, ar_name, ar_len);
   if (!name) {
@@ -451,12 +492,20 @@ int Process::handle_fstatat(const int fd, const char * const ar_name, const size
                             const int flags, const mode_t st_mode, const off_t st_size,
                             const int error) {
   TRACKX(FB_DEBUG_PROC, 1, 1, Process, this,
-         "fd=%d, ar_name=%s, flags=%d, st_mode=%d, st_size=%ld, error=%d",
+         "fd=%d, ar_name=%s, flags=%d, st_mode=%d, st_size=%" PRIoff ", error=%d",
          fd, D(ar_name), flags, st_mode, st_size, error);
 
   const FileName *name;
 
-  if (ar_name == nullptr || (ar_name[0] == '\0' && (flags & AT_EMPTY_PATH))) {
+#ifndef AT_EMPTY_PATH
+  (void)flags;
+#endif
+
+  if (ar_name == nullptr
+#ifdef AT_EMPTY_PATH
+      || (ar_name[0] == '\0' && (flags & AT_EMPTY_PATH))
+#endif
+      ) {
     /* Operating on an opened fd, i.e. fstat() or fstatat("", AT_EMPTY_PATH). */
     FileFD *file_fd = get_fd(fd);
     if (!file_fd) {
@@ -503,9 +552,13 @@ int Process::handle_statfs(const char * const a_name, const size_t length,
 
   if (a_name == nullptr) {
     /* Operating on an opened fd, i.e. fstatfs(). */
-    exec_point()->disable_shortcutting_bubble_up(
-        "fstatfs() family operating on fds is not supported");
-    return -1;
+    if (quirks & FB_QUIRK_IGNORE_STATFS) {
+      return 0;
+    } else {
+      exec_point()->disable_shortcutting_bubble_up(
+          "fstatfs() family operating on fds is not supported");
+      return -1;
+    }
   }
   /* Operating on a file reached by its name, made absolute by the interceptor. */
 #ifdef FB_EXTRA_DEBUG
@@ -517,7 +570,7 @@ int Process::handle_statfs(const char * const a_name, const size_t length,
     if (!exec_point()->register_file_usage_update(name, update)) {
       exec_point()->disable_shortcutting_bubble_up("Could not register failed statfs()", *name);
     }
-  } else {
+  } else if (!(quirks & FB_QUIRK_IGNORE_STATFS)) {
     // TODO(rbalint) add more supported cases
     exec_point()->disable_shortcutting_bubble_up(
         "Successful statfs() calls are not supported.");
@@ -643,7 +696,15 @@ int Process::handle_fchmodat(const int fd, const char * const ar_name, const siz
    * Let's just go on assuming it's supported to make our code consistent with fstatat(), future
    * fchownat() and possibly some other methods too. */
 
-  if (ar_name == nullptr || (ar_name[0] == '\0' && (flags & AT_EMPTY_PATH))) {
+#ifndef AT_EMPTY_PATH
+  (void)flags;
+#endif
+
+  if (ar_name == nullptr
+#ifdef AT_EMPTY_PATH
+      || (ar_name[0] == '\0' && (flags & AT_EMPTY_PATH))
+#endif
+      ) {
     /* Operating on an opened fd, i.e. fchmod() or fchmodat("", AT_EMPTY_PATH). */
     FileFD *file_fd = get_fd(fd);
     if (!file_fd) {
@@ -702,6 +763,7 @@ int Process::handle_fchmodat(const int fd, const char * const ar_name, const siz
   return 0;
 }
 
+#ifdef __linux__
 int Process::handle_memfd_create(const int flags, const int fd) {
   TRACKX(FB_DEBUG_PROC, 1, 1, Process, this, "flags=%d, fd=%d", flags, fd);
   add_filefd(fd, std::make_shared<FileFD>(fd, (flags & MFD_CLOEXEC) ? O_CLOEXEC : 0,
@@ -716,12 +778,14 @@ int Process::handle_timerfd_create(const int flags, const int fd) {
   return 0;
 }
 
+#ifdef __linux__
 int Process::handle_epoll_create(const int flags, const int fd) {
   TRACKX(FB_DEBUG_PROC, 1, 1, Process, this, "flags=%d, fd=%d", flags, fd);
   add_filefd(fd, std::make_shared<FileFD>(fd, (flags & EPOLL_CLOEXEC) ? O_CLOEXEC : 0,
                                           FD_SPECIAL, this));
   return 0;
 }
+#endif
 
 int Process::handle_eventfd(const int flags, const int fd) {
   TRACKX(FB_DEBUG_PROC, 1, 1, Process, this, "flags=%d, fd=%d", flags, fd);
@@ -742,6 +806,7 @@ int Process::handle_signalfd(const int oldfd, const int flags, const int newfd)
   }
   return 0;
 }
+#endif
 
 int Process::handle_rmdir(const char * const ar_name, const size_t ar_name_len, const int error,
                           const bool pre_open_sent) {
@@ -788,12 +853,12 @@ void Process::handle_pipe_request(const int flags, const int fd_conn) {
 
   FBBCOMM_Builder_pipe_created response;
 
-  if (pipe2(up, flags) < 0) {
+  if (fb_pipe2(up, flags) < 0) {
     response.set_error_no(errno);
     send_fbb(fd_conn, 0, reinterpret_cast<FBBCOMM_Builder *>(&response));
     return;
   }
-  if (pipe2(down, flags) < 0) {
+  if (fb_pipe2(down, flags) < 0) {
     response.set_error_no(errno);
     send_fbb(fd_conn, 0, reinterpret_cast<FBBCOMM_Builder *>(&response));
     close(up[0]);
@@ -906,17 +971,38 @@ void Process::handle_socket(const int domain, const int type, const int protocol
       handle_close(ret);
     }
     const int flags = O_RDWR
+#ifdef SOCK_CLOEXEC
         | ((type & SOCK_CLOEXEC) ? O_CLOEXEC : 0)
-        | ((type & SOCK_NONBLOCK) ? O_NONBLOCK : 0);
+#endif
+#ifdef SOCK_NONBLOCK
+        | ((type & SOCK_NONBLOCK) ? O_NONBLOCK : 0)
+#endif
+      | 0;
     add_filefd(ret, std::make_shared<FileFD>(ret, flags, FD_SPECIAL, this));
+#if defined(SOCK_CLOEXEC) && defined(SOCK_NONBLOCK)
     switch (type & ~(SOCK_CLOEXEC | SOCK_NONBLOCK)) {
+#elif defined(SOCK_CLOEXEC)
+#error ""
+#elif defined(SOCK_NONBLOCK)
+#else
+    switch (type) {
+#endif
+#if defined(SOCK_DGRAM)
       case SOCK_DGRAM:
+#endif
+#if defined(SOCK_RAW)
       case SOCK_RAW:
-      case SOCK_PACKET: {
+#endif
+#if defined(SOCK_PACKET)
+      case SOCK_PACKET:
+#endif
+#if defined(SOCK_DGRAM) || defined(SOCK_RAW) || defined(SOCK_PACKET)
+        {
         exec_point()->disable_shortcutting_bubble_up(
             "SOCK_DGRAM, SOCK_RAW and SOCK_PACKET sockets are not supported");
         break;
       }
+#endif
       default:
         /* Other socket types are OK since they require connect/bind/listen operations before
          * data can be exchanged with other processes and those disable shortcutting. */
@@ -939,6 +1025,9 @@ void Process::handle_socketpair(const int domain, const int type, const int prot
    * (e.g. memfd). */
   (void)domain;
   (void)protocol;
+#if !defined(SOCK_CLOEXEC) && !defined(SOCK_NONBLOCK)
+  (void)type;
+#endif
   if (!error) {
     for (const int fd : {fd0, fd1}) {
       if (get_fd(fd)) {
@@ -949,8 +1038,13 @@ void Process::handle_socketpair(const int domain, const int type, const int prot
         handle_close(fd);
       }
       const int flags = O_RDWR
+#ifdef SOCK_CLOEXEC
           | ((type & SOCK_CLOEXEC) ? O_CLOEXEC : 0)
-          | ((type & SOCK_NONBLOCK) ? O_NONBLOCK : 0);
+#endif
+#ifdef SOCK_NONBLOCK
+          | ((type & SOCK_NONBLOCK) ? O_NONBLOCK : 0)
+#endif
+        | 0;
       add_filefd(fd, std::make_shared<FileFD>(fd, flags, FD_SPECIAL, this));
     }
   } else {
@@ -1197,6 +1291,12 @@ void Process::handle_read_from_inherited(const int fd, const bool is_pread) {
         "Process successfully read from fd which is known to be closed, which means interception"
         " missed at least one open()", fd);
     return;
+  } else if (file_fd->type() == FD_PIPE_IN
+             && file_fd->pipe() && file_fd->fd() == exec_point()->jobserver_fd_r()) {
+    // TODO(rbalint) add further check that the process chain did not reopen the fd breaking the
+    // connection to the jobserver
+    /* It is OK to read from the jobserver. It should not affect the build results. */
+    return;
   }
 
   const ExecedProcess *file_fd_opened_by_exec_point =
@@ -1225,6 +1325,12 @@ void Process::handle_write_to_inherited(const int fd, const bool is_pwrite) {
         "Process successfully wrote to fd which is known to be closed, which means interception"
         " missed at least one open()", fd);
     return;
+  } else if (file_fd->type() == FD_PIPE_OUT
+             && file_fd->pipe() && file_fd->fd() == exec_point()->jobserver_fd_w()) {
+    // TODO(rbalint) add further check that the process chain did not reopen the fd breaking the
+    // connection to the jobserver
+    /* It is OK to write to the jobserver, but jobserver communication should not be cached. */
+    return;
   }
 
   const ExecedProcess *file_fd_opened_by_exec_point =
@@ -1285,7 +1391,7 @@ void Process::handle_seek_in_inherited(const int fd, const bool modify_offset) {
 }
 
 void Process::handle_inherited_fd_offset(const int fd, const int64_t offset) {
-  TRACKX(FB_DEBUG_PROC, 1, 1, Process, this, "fd=%d, offset==%ld", fd, offset);
+  TRACKX(FB_DEBUG_PROC, 1, 1, Process, this, "fd=%d, offset==%" PRId64, fd, offset);
   FileFD *file_fd = exec_point()->get_fd(fd);
   if (!file_fd) {
     exec_point()->disable_shortcutting_bubble_up(
@@ -1413,6 +1519,22 @@ const FileName* Process::get_absolute(const int dirfd, const char * const name,
   }
 }
 
+static bool argv_match_from_offset(const std::vector<std::string>& actual,
+                                   const int actual_offset,
+                                   const std::vector<std::string>& expected,
+                                   const int expected_offset) {
+  /* For the remaining parameters exact match is needed. */
+  if (actual.size() - actual_offset != expected.size() - expected_offset) {
+    return false;
+  }
+  for (unsigned int i = expected_offset; i < expected.size(); i++) {
+    if (actual[actual_offset + i - expected_offset] != expected[expected_offset]) {
+      return false;
+    }
+  }
+  return true;
+}
+
 static bool argv_matches_expectation(const std::vector<std::string>& actual,
                                      const std::vector<std::string>& expected) {
   /* When launching ["foo", "arg1"], the new process might be something like
@@ -1425,48 +1547,51 @@ static bool argv_matches_expectation(const std::vector<std::string>& actual,
    * A single interpreter might only add one additional command line parameter, but there can be
    * a chain of interpreters, so allow for arbitrarily many prepended parameters.
    */
-  if (actual.size() < expected.size()) {
-    return false;
-  }
 
   if (actual.size() == expected.size()) {
     /* If the length is the same, exact match is required. */
     return actual == expected;
   }
 
-  /* If the length grew, check the expected arg0. */
-  int offset = actual.size() - expected.size();
-  if (expected[0].find('/') != std::string::npos) {
-    /* If it contained a slash, exact match is needed. */
-    if (actual[offset] != expected[0]) {
-      return false;
+  if (actual.size() > expected.size()) {
+    /* If the length grew, check the expected arg0. */
+    int offset = actual.size() - expected.size();
+    if (expected[0].find('/') != std::string::npos) {
+      /* If it contained a slash, exact match is needed. */
+      if (actual[offset] != expected[0]) {
+        return false;
+      }
+    } else {
+      /* If it didn't contain a slash, it needed to get prefixed with some path. */
+      int expected_arg0_len = expected[0].length();
+      int actual_arg0_len = actual[offset].length();
+      /* Needs to be at least 1 longer. */
+      if (actual_arg0_len <= expected_arg0_len) {
+        return false;
+      }
+      /* See if the preceding character is a '/'. */
+      if (actual[offset][actual_arg0_len - expected_arg0_len - 1] != '/') {
+        return false;
+      }
+      /* See if the stuff after the '/' is the same. */
+      if (actual[offset].compare(actual_arg0_len - expected_arg0_len,
+                                 expected_arg0_len, expected[0]) != 0) {
+        return false;
+      }
     }
-  } else {
-    /* If it didn't contain a slash, it needed to get prefixed with some path. */
-    int expected_arg0_len = expected[0].length();
-    int actual_arg0_len = actual[offset].length();
-    /* Needs to be at least 1 longer. */
-    if (actual_arg0_len <= expected_arg0_len) {
-      return false;
-    }
-    /* See if the preceding character is a '/'. */
-    if (actual[offset][actual_arg0_len - expected_arg0_len - 1] != '/') {
-      return false;
-    }
-    /* See if the stuff after the '/' is the same. */
-    if (actual[offset].compare(actual_arg0_len - expected_arg0_len,
-                               expected_arg0_len, expected[0]) != 0) {
-      return false;
-    }
-  }
 
-  /* For the remaining parameters exact match is needed. */
-  for (unsigned int i = 1; i < expected.size(); i++) {
-    if (actual[offset + i] != expected[i]) {
-      return false;
+    return argv_match_from_offset(actual, offset + 1, expected, 1);
+  } else {
+    /* Actual argv is shorter than expected. */
+#if !FB_GLIBC_PREREQ (2, 38)
+    /* Older libc-s does not pass "--" between "sh -c" and system() and popen() argument. */
+    if (expected.size() > 2 && expected[2] == "--" && actual.size() > 2 && actual[2] != "--" &&
+        expected[0] == "sh" && expected[1] == "-c" && actual[0] == "sh" && actual[1] == "-c") {
+      return argv_match_from_offset(actual, 3, expected, 4);
     }
+#endif
+    return false;
   }
-  return true;
 }
 
 std::vector<std::shared_ptr<FileFD>>*
