@@ -275,6 +275,21 @@ class Process {
                      const int oldfd, const int fd, const int error = 0,
                      int fd_conn = -1, int ack_num = 0, bool pre_open_sent = false);
 
+  /**
+   * Handle dlopen() in the monitored process
+   * @param absolute_filename absolute file name
+   * @param absolute_filename_len length of absolute_filename
+   * @param looked_up_filename file name passed to dlopen()
+   * @param looked_up_filename_len length of looked_up_filename
+   * @param error true when dlopen failed to open the file
+   * @param fd_conn fd to send ACK on when needed
+   * @param ack_num ACK number to send or 0 if sending ACK is not needed
+   */
+  int handle_dlopen(const char * const absolute_filename,
+                    const size_t absolute_filename_len,
+                    const char * const looked_up_filename,
+                    const size_t looked_up_filename_len,
+                    const bool error, int fd_conn, int ack_num);
   /**
    * Handle file closure in the monitored process
    * @param fd file descriptor to close
@@ -368,6 +383,7 @@ class Process {
   int handle_fchmodat(const int fd, const char * const name, const size_t name_len,
                       const mode_t mode, const int flags, const int error = 0);
 
+#ifdef __linux__
   /**
    * Handle memfd_create in the monitored process
    */
@@ -395,6 +411,7 @@ class Process {
    * @param newfd new fd returned by signalfd()
    */
   int handle_signalfd(const int oldfd, const int flags, const int newfd);
+#endif
 
   /**
    * Handle rmdir in the monitored process
