@@ -41,6 +41,35 @@ namespace firebuild {
  * by C. */
 tsl::hopscotch_map<std::string, cmd_prof> cmd_profs {};
 
+/**
+ * Index of each used file in the JavaScript files[] array.
+ */
+tsl::hopscotch_map<const FileName*, int> used_files_index_map {};
+
+
+struct string_vector_ptr_hash {
+  size_t operator()(const std::vector<std::string>* v) const {
+    std::hash<std::string> hasher;
+    std::size_t seed = 0;
+    for (auto const & s : *v) {
+      seed ^= hasher(s) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
+    }
+    return seed;
+  }
+};
+
+struct string_vector_ptr_eq {
+  bool operator() (const std::vector<std::string>* a, const std::vector<std::string>* b) const {
+    return *a == *b;
+  }
+};
+
+/**
+ * Index of each used environment in the JavaScript envs[] array.
+ */
+tsl::hopscotch_map<const std::vector<std::string>*, int, struct string_vector_ptr_hash,
+                   struct string_vector_ptr_eq> used_envs_index_map {};
+
 /**
  * Escape std::string for JavaScript
  * from http://stackoverflow.com/questions/7724448/simple-json-string-escape-for-c
@@ -68,8 +97,8 @@ static const char* full_relative_path_or_basename(const char *name) {
   return name_last_slash && path_is_absolute(name) ? name_last_slash + 1 : name;
 }
 
-static void fprintf_escaped_file(FILE* stream, const file_file_usage& ffu) {
-  fprintf(stream, "\"%s\",", escapeJsonString(ffu.file->to_string()).c_str());
+static void fprintf_ffu_file(FILE* stream, const file_file_usage& ffu) {
+  fprintf(stream, "files[%d],", used_files_index_map[ffu.file]);
 }
 
 static void export2js(const ExecedProcess* proc, const unsigned int level,
@@ -89,6 +118,10 @@ static void export2js(const ExecedProcess* proc, const unsigned int level,
   if (proc->was_shortcut()) {
     fprintf(stream, "%s was_shortcut: true,\n", indent);
   }
+  if (proc->shortcut_result()) {
+    fprintf(stream, "%s sc_result: \"%s\",\n",
+            indent, escapeJsonString(proc->shortcut_result()).c_str());
+  }
   if (!proc->can_shortcut()) {
     fprintf(stream, "%s cant_sc_reason: \"%s\",\n",
             indent, escapeJsonString(proc->cant_shortcut_reason()).c_str());
@@ -103,11 +136,7 @@ static void export2js(const ExecedProcess* proc, const unsigned int level,
   }
   fprintf(stream, "],\n");
 
-  fprintf(stream, "%s env: [", indent);
-  for (auto& env : proc->env_vars()) {
-    fprintf(stream, "\"%s\",", escapeJsonString(env).c_str());
-  }
-  fprintf(stream, "],\n");
+  fprintf(stream, "%s env: envs[%d],\n", indent, used_envs_index_map[&proc->env_vars()]);
 
   fprintf(stream, "%s libs: [", indent);
   for (auto& lib : proc->libs()) {
@@ -130,7 +159,9 @@ static void export2js(const ExecedProcess* proc, const unsigned int level,
   /* sort files before printing */
   std::vector<file_file_usage> ordered_file_usages;
   for (auto& pair : proc->file_usages()) {
-    ordered_file_usages.push_back({pair.first, pair.second});
+    if (!pair.second->propagated()) {
+      ordered_file_usages.push_back({pair.first, pair.second});
+    }
   }
   std::sort(ordered_file_usages.begin(), ordered_file_usages.end(), file_file_usage_cmp);
 
@@ -138,7 +169,7 @@ static void export2js(const ExecedProcess* proc, const unsigned int level,
   for (auto& ffu : ordered_file_usages) {
     bool isreg_with_hash = ffu.usage->initial_type() == ISREG && ffu.usage->initial_hash_known();
     if (!isreg_with_hash && ffu.usage->written()) {
-      fprintf_escaped_file(stream, ffu);
+      fprintf_ffu_file(stream, ffu);
     }
   }
   fprintf(stream, "],\n");
@@ -147,7 +178,7 @@ static void export2js(const ExecedProcess* proc, const unsigned int level,
   for (auto& ffu : ordered_file_usages) {
     bool isreg_with_hash = ffu.usage->initial_type() == ISREG && ffu.usage->initial_hash_known();
     if (isreg_with_hash && ffu.usage->written()) {
-      fprintf_escaped_file(stream, ffu);
+      fprintf_ffu_file(stream, ffu);
     }
   }
   fprintf(stream, "],\n");
@@ -156,7 +187,7 @@ static void export2js(const ExecedProcess* proc, const unsigned int level,
   for (auto& ffu : ordered_file_usages) {
     bool isreg_with_hash = ffu.usage->initial_type() == ISREG && ffu.usage->initial_hash_known();
     if (isreg_with_hash && !ffu.usage->written()) {
-      fprintf_escaped_file(stream, ffu);
+      fprintf_ffu_file(stream, ffu);
     }
   }
   fprintf(stream, "],\n");
@@ -164,7 +195,7 @@ static void export2js(const ExecedProcess* proc, const unsigned int level,
   fprintf(stream, "%s fnotf: [", indent);
   for (auto& ffu : ordered_file_usages) {
     if (ffu.usage->initial_type() == NOTEXIST) {
-      fprintf_escaped_file(stream, ffu);
+      fprintf_ffu_file(stream, ffu);
     }
   }
   fprintf(stream, "],\n");
@@ -221,6 +252,54 @@ static void export2js(ProcessTree* proc_tree, FILE * stream) {
   }
 }
 
+static void collect_used_files_and_envs(const Process &p,
+                                        tsl::hopscotch_set<const FileName*> *used_files,
+                                        tsl::hopscotch_set<const std::vector<std::string>*,
+                                        string_vector_ptr_hash,
+                                        string_vector_ptr_eq> *envs) {
+  if (p.exec_child() != NULL) {
+    ExecedProcess *exec_child = static_cast<ExecedProcess*>(p.exec_child());
+    for (const auto& pair : exec_child->file_usages()) {
+      if (!pair.second->propagated()) { /* Save time by not processing propagated ones. */
+        used_files->insert(pair.first);
+      }
+    }
+    envs->insert(&exec_child->env_vars());
+    collect_used_files_and_envs(*exec_child, used_files, envs);
+  }
+  for (auto& fork_child : p.fork_children()) {
+    collect_used_files_and_envs(*fork_child, used_files, envs);
+  }
+}
+
+void fprint_collected_files(FILE* stream,
+                            const tsl::hopscotch_set<const FileName*>& used_files_set) {
+  int index = 0;
+  fprintf(stream, "files = [\n");
+  for (const FileName* filename : used_files_set) {
+    fprintf(stream, "  \"%s\", // files[%d]\n",
+            escapeJsonString(filename->to_string()).c_str(), index);
+    used_files_index_map.insert({filename, index++});
+  }
+  fprintf(stream, "];\n");
+}
+
+void fprint_collected_envs(
+    FILE* stream, const tsl::hopscotch_set<const std::vector<std::string>*, string_vector_ptr_hash,
+    string_vector_ptr_eq>& used_envs_set) {
+  int index = 0;
+  fprintf(stream, "envs = [\n");
+  for (const std::vector<std::string>* env : used_envs_set) {
+    fprintf(stream, "  [");
+    for (const std::string& env_var : *env) {
+      fprintf(stream, "\"%s\",", escapeJsonString(env_var).c_str());
+    }
+    fprintf(stream, "], // envs[%d]\n", index);
+    used_envs_index_map.insert({env, index++});
+  }
+  fprintf(stream, "];\n");
+}
+
 static void profile_collect_cmds(const Process &p,
                                  tsl::hopscotch_map<std::string, subcmd_prof> *cmds,
                                  std::set<std::string> *ancestors) {
@@ -463,6 +542,12 @@ void Report::write(const std::string &html_filename, const std::string &datadir)
       }
     } else if (strstr(line, tree_filename) != NULL) {
       fprintf(dst_file, "<script type=\"text/javascript\">\n");
+      tsl::hopscotch_set<const FileName*> used_files_set;
+      tsl::hopscotch_set<const std::vector<std::string>*, string_vector_ptr_hash,
+                         string_vector_ptr_eq> used_envs_set;
+      collect_used_files_and_envs(*proc_tree->root(), &used_files_set, &used_envs_set);
+      fprint_collected_files(dst_file, used_files_set);
+      fprint_collected_envs(dst_file, used_envs_set);
       export2js(proc_tree, dst_file);
       fprintf(dst_file, "    </script>\n");
     } else if (strstr(line, svg_filename) != NULL) {
