@@ -22,9 +22,16 @@
 #include <unistd.h>
 #include <errno.h>
 #include <fcntl.h>
+#ifdef __APPLE__
+#include "libproc.h"
+#endif
+#ifdef __linux__
 #include <link.h>
+#endif
 #include <pthread.h>
+#ifdef __linux__
 #include <sys/auxv.h>
+#endif
 #include <sys/un.h>
 #include <sys/resource.h>
 #include <spawn.h>
@@ -42,7 +49,7 @@
 #define VDSO_NAME "linux-vdso.so.1"
 #endif
 
-static void fb_ic_load_constructor(int argc, char **argv) __attribute__((constructor));
+static void fb_ic_init_constructor(int argc, char **argv) __attribute__((constructor));
 static void fb_ic_cleanup() __attribute__((destructor));
 
 fd_state ic_fd_states[IC_FD_STATES_SIZE];
@@ -53,32 +60,34 @@ pthread_mutex_t ic_system_popen_lock = PTHREAD_MUTEX_INITIALIZER;
 
 pthread_mutex_t ic_global_lock = PTHREAD_MUTEX_INITIALIZER;
 
-char fb_conn_string[IC_PATH_BUFSIZE] = {'\0'};
+char fb_conn_string[FB_PATH_BUFSIZE] = {'\0'};
 size_t fb_conn_string_len = 0;
 
 int fb_sv_conn = -1;
 
-int (*ic_pthread_sigmask)(int, const sigset_t *, sigset_t *);
-
 bool ic_called_syscall[IC_CALLED_SYSCALL_SIZE] = {0};
 
-/** Control for running the initialization exactly once */
-pthread_once_t ic_init_control = PTHREAD_ONCE_INIT;
-
+bool ic_init_started = false;
 bool ic_init_done = false;
 
 /** System locations to not ask ACK for when opening them, as set in the environment variable. */
-char system_locations_env_buf[4096];
+char read_only_locations_env_buf[4096];
 
 /** Ignore locations to not ask ACK for when opening them, as set in the environment variable. */
 char ignore_locations_env_buf[4096];
 
-STATIC_CSTRING_VIEW_ARRAY(system_locations, 32);
+/**
+ * Jobserver users for which the jobserver fds have to be detected, as set in the environment
+ * variable. */
+char jobserver_users_env_buf[4096];
+
+STATIC_CSTRING_VIEW_ARRAY(read_only_locations, 32);
 STATIC_CSTRING_VIEW_ARRAY(ignore_locations, 32);
+STATIC_CSTRING_VIEW_ARRAY(jobserver_users, 8);
 
 bool intercepting_enabled = true;
 
-char ic_cwd[IC_PATH_BUFSIZE] = {0};
+char ic_cwd[FB_PATH_BUFSIZE] = {0};
 size_t ic_cwd_len = 0;
 
 /** Program's argc and argv. */
@@ -93,6 +102,11 @@ __thread bool thread_has_global_lock = false;
 __thread sig_atomic_t thread_signal_handler_running_depth = 0;
 __thread sig_atomic_t thread_libc_nesting_depth = 0;
 __thread uint64_t thread_delayed_signals_bitmap = 0;
+#ifdef __APPLE__
+/* OS X does not support RT signals https://flylib.com/books/en/3.126.1.110/1/ ,
+ * but we can handle 64 signals safely. */
+#define SIGRTMAX ((int)sizeof(thread_delayed_signals_bitmap) * 8)
+#endif
 
 void (*orig_signal_handlers[IC_WRAP_SIGRTMAX])(void) = {NULL};
 
@@ -169,6 +183,30 @@ void thread_raise_delayed_signals() {
   }
 }
 
+int ic_pthread_sigmask(int how, const sigset_t *set, sigset_t *oldset) {
+  /* pthread_sigmask() is only available if we're linked against libpthread.
+   * Otherwise use the single-threaded sigprocmask(). */
+#if defined(__APPLE__) || FB_GLIBC_PREREQ(2, 34)
+  return pthread_sigmask(how, set, oldset);
+#else
+  static int (*ic_orig_pthread_sigmask)(int, const sigset_t *, sigset_t *);
+  static bool tried_dlsym = false;
+
+  if (ic_orig_pthread_sigmask) {
+    return ic_orig_pthread_sigmask(how, set, oldset);
+  } else {
+    if (!tried_dlsym) {
+      ic_orig_pthread_sigmask = dlsym(RTLD_NEXT, "pthread_sigmask");
+      tried_dlsym = true;
+      /* Try again with possibly resolved symbol. */
+      return ic_pthread_sigmask(how, set, oldset);
+    } else {
+      return sigprocmask(how, set, oldset);
+    }
+  }
+#endif
+}
+
 void grab_global_lock(bool *i_locked, const char * const function_name) {
   thread_signal_danger_zone_enter();
 
@@ -216,7 +254,7 @@ void release_global_lock() {
 /** debugging flags */
 int32_t debug_flags = 0;
 
-char env_ld_library_path[IC_PATH_BUFSIZE] = {0};
+char env_ld_library_path[FB_PATH_BUFSIZE] = {0};
 
 bool insert_trace_markers = false;
 
@@ -363,7 +401,9 @@ void send_pre_open_without_ack_request(const int dirfd, const char* pathname) {
 bool maybe_send_pre_open(const int dirfd, const char* pathname, int flags) {
   if (pathname && is_write(flags) && (flags & O_TRUNC)
       && !(flags & (O_EXCL | O_DIRECTORY))
+#ifdef O_TMPFILE
       && (flags & O_TMPFILE) != O_TMPFILE
+#endif
       && !is_path_at_locations(pathname, -1, &ignore_locations)) {
     send_pre_open(dirfd, pathname);
     return true;
@@ -527,29 +567,29 @@ static int cmpstringpp(const void *p1, const void *p2) {
   return strcmp(* (char * const *) p1, * (char * const *) p2);
 }
 
-/** Store file locations for which files open() does not need an ACK. */
-static void store_locations(const char* env_var, cstring_view_array *locations,
-                            char * locations_env_buf, size_t buffer_size) {
-  char* env_locations = getenv(env_var);
-  if (env_locations) {
-    strncpy(locations_env_buf, env_locations, buffer_size);
-    const size_t env_locations_len = strlen(env_locations);
-    if (env_locations_len + 1 > buffer_size) {
-      /* Trim to the fitting parts. The locations are used only for improving
+/** Store from entries from environment variable. */
+static void store_entries(const char* env_var, cstring_view_array *entries,
+                          char * entries_env_buf, size_t buffer_size) {
+  char* env_entries = getenv(env_var);
+  if (env_entries) {
+    strncpy(entries_env_buf, env_entries, buffer_size);
+    const size_t env_entries_len = strlen(env_entries);
+    if (env_entries_len + 1 > buffer_size) {
+      /* Trim to the fitting parts. The entries are used only for improving
        * performance and the space is allocated statically. */
-      locations_env_buf[buffer_size - 1] = '\0';
-      char * last_separator = strrchr(locations_env_buf, ':');
+      entries_env_buf[buffer_size - 1] = '\0';
+      char * last_separator = strrchr(entries_env_buf, ':');
       if (!last_separator) {
         /* This is a quite long single path that may be incomplete, thus ignore it. */
-        locations_env_buf[0] = '\0';
+        entries_env_buf[0] = '\0';
       } else {
         /* Drop the possibly incomplete path after the last separator.*/
         *last_separator = '\0';
       }
     }
-    char *prefix = locations_env_buf;
-    /* Process all locations that fit location without reallocation. */
-    while (prefix && !is_cstring_view_array_full(locations)) {
+    char *prefix = entries_env_buf;
+    /* Process all entries that fit location without reallocation. */
+    while (prefix && !is_cstring_view_array_full(entries)) {
       char *next_prefix = strchr(prefix, ':');
       if (next_prefix) {
         *next_prefix = '\0';
@@ -557,13 +597,33 @@ static void store_locations(const char* env_var, cstring_view_array *locations,
       }
       /* Skip "". */
       if (*prefix != '\0') {
-        cstring_view_array_append_noalloc(locations, prefix);
+        cstring_view_array_append_noalloc(entries, prefix);
         prefix = next_prefix;
       }
     }
   }
 }
 
+#ifdef __APPLE__
+static void collect_shared_libs(cstring_view_array* libs, char *canonized_libs) {
+  /* Skip first image because it is the binary itself. */
+  for (int32_t i = _dyld_image_count() - 1; i > 1 ; i--) {
+    const char *image_name = _dyld_get_image_name(i);
+    const size_t len = strlen(image_name);
+    if (is_canonical(image_name, len)) {
+      assert(!is_cstring_view_array_full(libs));
+      cstring_view_array_append_noalloc(libs, (/* not const */ char *)image_name);
+    } else {
+      char *canonical_name = &canonized_libs[i * FB_PATH_BUFSIZE];
+      memcpy(canonical_name, image_name, len + 1);
+      make_canonical(canonical_name, len);
+      assert(!is_cstring_view_array_full(libs));
+      cstring_view_array_append_noalloc(libs, canonical_name);
+    }
+  }
+}
+
+#else
 static bool skip_shared_lib(const char *name, const size_t len) {
   if (name[0] == '\0') {
     /* FIXME does this really happen? */
@@ -593,7 +653,7 @@ typedef struct shared_libs_cb_data_ {
   int collectable_entries;
   /** Number of entries that are not in canonical form, thus need to be made canonical. */
   int not_canonical_entries;
-  /** Buffert to store canonized library names. Size is canonized_libs_size * IC_PATH_BUFSIZE. */
+  /** Buffert to store canonized library names. Size is canonized_libs_size * FB_PATH_BUFSIZE. */
   char *canonized_libs;
   /** Number of canonized names canonized_libs can store. */
   int canonized_libs_size;
@@ -624,7 +684,7 @@ static int shared_libs_cb(struct dl_phdr_info *info, const size_t size, void *da
     if (cb_data->canonized_libs_count < cb_data->canonized_libs_size) {
       /* The there is enough space for the new canonized entry. */
       char * canonical_name =
-          &cb_data->canonized_libs[cb_data->canonized_libs_count++ * IC_PATH_BUFSIZE];
+          &cb_data->canonized_libs[cb_data->canonized_libs_count++ * FB_PATH_BUFSIZE];
       memcpy(canonical_name, name, len + 1);
       make_canonical(canonical_name, len);
       if (!is_cstring_view_array_full(array)) {
@@ -634,6 +694,7 @@ static int shared_libs_cb(struct dl_phdr_info *info, const size_t size, void *da
   }
   return 0;
 }
+#endif
 
 void atfork_parent_handler(void) {
   /* The variable i_am_intercepting from the intercepted fork() is
@@ -750,12 +811,50 @@ void *pthread_start_routine_wrapper(void *routine_and_arg) {
   return (*start_routine)(arg);
 }
 
+/**
+ * Parses and returns GNU Make jobserver fds if they are present in makeflags.
+ * e.g. --jobserver-auth=R,W where ‘R’ and ‘W’ are non-negative integers representing fds
+ *
+ * @param[in] makeflags Make flags as set in environment's MAKEFLAGS
+ * @param[out] fd_r R fd
+ * @param[out] fd_w W fd
+ * @return true, if jobserver fds were set and parsed
+ */
+static bool extract_jobserver_fds(const char* makeflags_env, int *fd_r, int *fd_w) {
+  const char *makeflags = getenv(makeflags_env);
+  if (!makeflags) {
+    return false;
+  }
+  const char *needle = "--jobserver-auth=";
+  const char *jobserver_option = strstr(makeflags, needle);
+  if (!jobserver_option) {
+    needle = "--jobserver-fds=";
+    jobserver_option = strstr(makeflags, needle);
+  }
+  if (jobserver_option) {
+    if (sscanf(jobserver_option + strlen(needle), "%d,%d", fd_r, fd_w) == 2) {
+      return true;
+    }
+  }
+  return false;
+}
+
+
 /**
  * Set up a supervisor connection
  * @return fd of the connection
  */
 int fb_connect_supervisor() {
+#ifdef SOCK_CLOEXEC
   int conn = TEMP_FAILURE_RETRY(get_ic_orig_socket()(AF_UNIX, SOCK_STREAM | SOCK_CLOEXEC, 0));
+#else
+  int conn = TEMP_FAILURE_RETRY(get_ic_orig_socket()(AF_UNIX, SOCK_STREAM, 0));
+#ifndef NDEBUG
+  int fcntl_ret =
+#endif
+      TEMP_FAILURE_RETRY(get_ic_orig_fcntl()(conn, F_SETFD, FD_CLOEXEC));
+  assert(fcntl_ret != -1);
+#endif
   assert(conn != -1);
 
   struct sockaddr_un remote;
@@ -790,13 +889,16 @@ void fb_init_supervisor_conn() {
 /**
  * Detect main()'s argc and argv with heuristics.
  *
- * The reliable and portable initialization happens in fb_ic_load_constructor(), but in case
+ * The reliable and portable initialization happens in fb_ic_init_constructor(), but in case
  * an intercepted function is called before the constructor of this shared library, argc
  * and argv still needs to be reported to the supervisor in the first message.
  * The heuristics below works for most programs, but not for mpicc. Luckily when intercepting
  * mpicc the constructor is called first, thus this heuristics is not used.
  */
 static void init_argc_argv() {
+#ifdef __APPLE__
+  char** __environ = environ;
+#endif
   if (ic_argv == NULL) {
     char* arg = *(__environ - 2);
     unsigned long int argc_guess = 0;
@@ -814,19 +916,31 @@ static void init_argc_argv() {
 }
 
 /**
- * Initialize interceptor's data structures and sync with supervisor
+ * Initialize interceptor's data structures and sync with supervisor.
+ *
+ * Collect information about process the earliest possible, right
+ * when interceptor library loads or when the first interceped call happens
  */
-static void fb_ic_init() {
+void fb_ic_init() {
+  /* Run only once, at startup. */
+  if (ic_init_started) {
+    /* Should not be called recursively. */
+    assert(ic_init_done);
+    return;
+  }
+  ic_init_started = true;
   get_ic_orig_getrusage()(RUSAGE_SELF, &initial_rusage);
 
   if (getenv("FB_INSERT_TRACE_MARKERS") != NULL) {
     insert_trace_markers = true;
   }
 
-  store_locations("FB_SYSTEM_LOCATIONS", &system_locations, system_locations_env_buf,
-                  sizeof(system_locations_env_buf));
-  store_locations("FB_IGNORE_LOCATIONS", &ignore_locations, ignore_locations_env_buf,
-                  sizeof(ignore_locations_env_buf));
+  store_entries("FB_READ_ONLY_LOCATIONS", &read_only_locations, read_only_locations_env_buf,
+                sizeof(read_only_locations_env_buf));
+  store_entries("FB_IGNORE_LOCATIONS", &ignore_locations, ignore_locations_env_buf,
+                sizeof(ignore_locations_env_buf));
+  store_entries("FB_JOBSERVER_USERS", &jobserver_users, jobserver_users_env_buf,
+                sizeof(jobserver_users_env_buf));
 
 #ifndef __mips__
   /* We use an uint64_t as bitmap for delayed signals. Make sure it's okay.
@@ -893,7 +1007,11 @@ static void fb_ic_init() {
   fbbcomm_builder_scproc_query_set_umask(&ic_msg, initial_umask);
 
   /* make a sorted and filtered copy of env */
+#ifdef __APPLE__
+  char **env = environ;
+#else
   char **env = __environ;
+#endif
   int env_len = 0, env_copy_len = 0;
   for (char** cursor = env; *cursor != NULL; cursor++) {
     env_len++;
@@ -902,11 +1020,13 @@ static void fb_ic_init() {
 
   for (char** cursor = env; *cursor != NULL; cursor++) {
     const char *fb_socket = "FB_SOCKET=";
-    const char *fb_system_locations = "FB_SYSTEM_LOCATIONS=";
+    const char *fb_read_only_locations = "FB_READ_ONLY_LOCATIONS=";
     const char *fb_ignore_locations = "FB_IGNORE_LOCATIONS=";
+    const char *fb_jobserver_users = "FB_JOBSERVER_USERS=";
     if (strncmp(*cursor, fb_socket, strlen(fb_socket)) != 0 &&
-        strncmp(*cursor, fb_system_locations, strlen(fb_system_locations)) != 0 &&
-        strncmp(*cursor, fb_ignore_locations, strlen(fb_ignore_locations)) != 0) {
+        strncmp(*cursor, fb_read_only_locations, strlen(fb_read_only_locations)) != 0 &&
+        strncmp(*cursor, fb_ignore_locations, strlen(fb_ignore_locations)) != 0 &&
+        strncmp(*cursor, fb_jobserver_users, strlen(fb_jobserver_users)) != 0) {
       env_copy[env_copy_len++] = *cursor;
     }
   }
@@ -914,19 +1034,53 @@ static void fb_ic_init() {
   qsort(env_copy, env_copy_len, sizeof(env_copy[0]), cmpstringpp);
   fbbcomm_builder_scproc_query_set_env_var(&ic_msg, (const char **) env_copy);
 
+  const char* slash_pos = strrchr(ic_argv[0], '/');
+  const char* cmd_name = slash_pos ? slash_pos + 1 : ic_argv[0];
+  int jobserver_fds[] = {-1, -1};
+  if (is_in_sorted_cstring_view_array(cmd_name, strlen(cmd_name), &jobserver_users)) {
+    if (extract_jobserver_fds("CARGO_MAKEFLAGS", &jobserver_fds[0], &jobserver_fds[1])) {
+      fbbcomm_builder_scproc_query_set_jobserver_fds(&ic_msg, jobserver_fds, 2);
+    } else if (extract_jobserver_fds("MAKEFLAGS", &jobserver_fds[0], &jobserver_fds[1])) {
+      fbbcomm_builder_scproc_query_set_jobserver_fds(&ic_msg, jobserver_fds, 2);
+    }
+  }
+
   /* get full executable path
    * see http://stackoverflow.com/questions/1023306/finding-current-executables-path-without-proc-self-exe
    * and readlink(2) */
-  char linkname[IC_PATH_BUFSIZE];
-  ssize_t r;
-  r = get_ic_orig_readlink()("/proc/self/exe", linkname, IC_PATH_BUFSIZE - 1);
-  if (r > 0 && r < IC_PATH_BUFSIZE) {
+  char linkname[FB_PATH_BUFSIZE];
+#ifdef __APPLE__
+  uint32_t r = sizeof(linkname);
+  if (_NSGetExecutablePath(linkname, &r) == 0) {
+    r = strlen(linkname);
+  } else {
+    /* A bigger buffer is needed. */
+    char* linkname2 = alloca(++r);
+    if (_NSGetExecutablePath(linkname2, &r) != 0) {
+      assert(0 && "Could not get the executable path even with the buffer "
+             "that should have been enough.");
+      r = 0;
+    } else {
+      r = strlen(linkname2);
+      fbbcomm_builder_scproc_query_set_executable_with_length(&ic_msg, linkname2, r);
+    }
+  }
+#else
+  ssize_t r = get_ic_orig_readlink()("/proc/self/exe", linkname, FB_PATH_BUFSIZE - 1);
+#endif
+  if (r > 0 && r < FB_PATH_BUFSIZE) {
     linkname[r] = '\0';
     fbbcomm_builder_scproc_query_set_executable_with_length(&ic_msg, linkname, r);
   }
 
+#ifdef __APPLE__
+  char original_executed_path[PROC_PIDPATHINFO_MAXSIZE];
+  if (proc_pidpath(getpid(), original_executed_path, sizeof(original_executed_path)) != -1
+#else
   const char *original_executed_path = (const char*)getauxval(AT_EXECFN);
-  if (original_executed_path && strcmp(original_executed_path, linkname) != 0) {
+  if (original_executed_path
+#endif
+      && strcmp(original_executed_path, linkname) != 0) {
     /* The macro relies on the field name matching the variable name. */
     const char *executed_path = original_executed_path;
     BUILDER_SET_ABSOLUTE_CANONICAL(scproc_query, executed_path);
@@ -937,16 +1091,23 @@ static void fb_ic_init() {
   }
 
   /* list loaded shared libs */
+#ifdef __APPLE__
+  const int image_count = _dyld_image_count();
+  cstring_view *libs_ptrs = alloca((image_count + 1) * sizeof(cstring_view));
+  cstring_view_array libs = {libs_ptrs, 0, image_count + 1};
+  char *canonized_libs = alloca(image_count * FB_PATH_BUFSIZE);
+  collect_shared_libs(&libs, canonized_libs);
+#else
   STATIC_CSTRING_VIEW_ARRAY(libs, 64);
   int canonized_libs_size = 8;
-  char *canonized_libs = alloca(canonized_libs_size * IC_PATH_BUFSIZE);
+  char *canonized_libs = alloca(canonized_libs_size * FB_PATH_BUFSIZE);
   shared_libs_cb_data_t cb_data = {&libs, 0, 0, canonized_libs, canonized_libs_size, 0};
   dl_iterate_phdr(shared_libs_cb, &cb_data);
   if (cb_data.collectable_entries > cb_data.array->len) {
     if (cb_data.not_canonical_entries > canonized_libs_size) {
       /* canonized_libs was not big enough. */
       canonized_libs_size = cb_data.not_canonical_entries;
-      canonized_libs = alloca(canonized_libs_size * IC_PATH_BUFSIZE);
+      canonized_libs = alloca(canonized_libs_size * FB_PATH_BUFSIZE);
     }
     /* The initially allocated space was not enough to collect all shared libs, trying again. */
     if (cb_data.collectable_entries > cb_data.array->size_alloc - 1) {
@@ -963,6 +1124,7 @@ static void fb_ic_init() {
     dl_iterate_phdr(shared_libs_cb, &cb_data2);
     assert(cb_data.collectable_entries == cb_data2.array->len);
   }
+#endif
   fbbcomm_builder_scproc_query_set_libs_cstring_views(&ic_msg, libs.p, libs.len);
 
   fb_send_msg(fb_sv_conn, &ic_msg, 0);
@@ -1025,8 +1187,12 @@ static void fb_ic_init() {
     }
 
     insert_debug_msg("exiting");
+#ifdef __APPLE__
+    _exit(fbbcomm_serialized_scproc_resp_get_exit_status(sv_msg));
+#else
     void(*orig_underscore_exit)(int) = (void(*)(int)) dlsym(RTLD_NEXT, "_exit");
     (*orig_underscore_exit)(fbbcomm_serialized_scproc_resp_get_exit_status(sv_msg));
+#endif
     assert(0 && "_exit() did not exit");
   }
 
@@ -1117,7 +1283,11 @@ static void fb_ic_init() {
     int fd = fbbcomm_serialized_scproc_resp_get_seekable_fds_at(sv_msg, i);
     int64_t size = fbbcomm_serialized_scproc_resp_get_seekable_fds_size_at(sv_msg, i);
     insert_debug_msg("get offset of fd");
+#ifdef __APPLE__
+    off_t offset = get_ic_orig_lseek()(fd, 0, SEEK_CUR);
+#else
     off64_t offset = get_ic_orig_lseek64()(fd, 0, SEEK_CUR);
+#endif
     if (offset != size) {
       FBBCOMM_Builder_inherited_fd_offset ic_msg;
       fbbcomm_builder_inherited_fd_offset_init(&ic_msg);
@@ -1127,41 +1297,16 @@ static void fb_ic_init() {
     }
   }
 
-  /* pthread_sigmask() is only available if we're linked against libpthread.
-   * Otherwise use the single-threaded sigprocmask(). */
-  ic_pthread_sigmask = dlsym(RTLD_NEXT, "pthread_sigmask");
-  if (!ic_pthread_sigmask) {
-    ic_pthread_sigmask = &sigprocmask;
-  }
-
   insert_debug_msg("initialization-end");
   thread_intercept_on = NULL;
   ic_init_done = true;
 }
 
-/**
- * Collect information about process the earliest possible, right
- * when interceptor library loads or when the first interceped call happens
- */
-void fb_ic_load() {
-  if (!ic_init_done) {
-    int (*orig_pthread_once)(pthread_once_t *, void (*)(void)) = dlsym(RTLD_NEXT, "pthread_once");
-    if (orig_pthread_once) {
-      /* Symbol found means that we are linked to libpthread. Use its method to guarantee that we
-       * initialize exactly once. */
-      (*orig_pthread_once)(&ic_init_control, fb_ic_init);
-    } else  {
-      /* Symbol not found means that we are not linked to libpthread, i.e. we're single threaded. */
-      fb_ic_init();
-    }
-  }
-}
-
-static void fb_ic_load_constructor(int argc, char **argv) {
-  if (!ic_init_done) {
+static void fb_ic_init_constructor(int argc, char **argv) {
+  if (!ic_init_started) {
     ic_argc = argc;
     ic_argv = argv;
-    fb_ic_load();
+    fb_ic_init();
   }
 }
 
