@@ -27,9 +27,14 @@
 #define _GNU_SOURCE 1
 #endif
 
+#ifdef __linux__
 #include <bits/types/FILE.h>
+#endif
+
 #include <dlfcn.h>
+#ifdef __linux__
 #include <link.h>
+#endif
 #include <pthread.h>
 #include <dirent.h>
 #include <signal.h>
@@ -110,11 +115,8 @@ extern struct rusage initial_rusage;
 /** Global lock for preventing parallel system and popen calls */
 extern pthread_mutex_t ic_system_popen_lock;
 
-/** buffer size for getcwd */
-#define IC_PATH_BUFSIZE 4096
-
 /** Current working directory as reported to the supervisor */
-extern char ic_cwd[IC_PATH_BUFSIZE];
+extern char ic_cwd[FB_PATH_BUFSIZE];
 extern size_t ic_cwd_len;
 
 /** Reset globally maintained information about intercepted functions */
@@ -161,7 +163,7 @@ void pre_clone_disable_interception(const int flags, bool *i_locked);
 int clone_trampoline(void *arg);
 
 /** Connection string to supervisor */
-extern char fb_conn_string[IC_PATH_BUFSIZE];
+extern char fb_conn_string[FB_PATH_BUFSIZE];
 
 /** Connection string length */
 extern size_t fb_conn_string_len;
@@ -170,9 +172,11 @@ extern size_t fb_conn_string_len;
 extern int fb_sv_conn;
 
 /** pthread_sigmask() if available (libpthread is loaded), otherwise sigprocmask() */
-extern int (*ic_pthread_sigmask)(int, const sigset_t *, sigset_t *);
+int ic_pthread_sigmask(int, const sigset_t *, sigset_t *);
 
-/** Fast check for whether interceptor init has been run */
+/** Fast check for whether interceptor init has been started. */
+extern bool ic_init_started;
+/** Fast check for whether interceptor init has finished. */
 extern bool ic_init_done;
 
 extern bool intercepting_enabled;
@@ -227,13 +231,13 @@ extern voidp_array *psfa_find(const posix_spawn_file_actions_t *p);
 extern voidp_set popened_streams;
 
 /** Initial LD_LIBRARY_PATH so that we can fix it up if needed */
-extern char env_ld_library_path[IC_PATH_BUFSIZE];
+extern char env_ld_library_path[FB_PATH_BUFSIZE];
 
 /** Insert marker open()-s for strace, ltrace, etc. */
 extern bool insert_trace_markers;
 
 /** System and ignore locations to not ask ACK for when opening them. */
-extern cstring_view_array system_locations;
+extern cstring_view_array read_only_locations;
 extern cstring_view_array ignore_locations;
 
 /** Insert debug message */
@@ -267,7 +271,7 @@ size_t make_canonical(char *path, size_t original_length);
 
 #ifdef FB_EXTRA_DEBUG
 static inline bool ic_cwd_ok() {
-  char buf[IC_PATH_BUFSIZE];
+  char buf[FB_PATH_BUFSIZE];
   /* getcwd() is not intercepted */
   char * getcwd_ret = getcwd(buf, sizeof(buf));
   assert(getcwd_ret);
@@ -436,7 +440,7 @@ void atfork_parent_handler(void);
  */
 void atfork_child_handler(void);
 
-extern void fb_ic_load();
+extern void fb_ic_init();
 extern void handle_exit();
 /**
  * A wrapper in front of the start_routine of a pthread_create(), inserting a useful trace marker.
